import warnings
from typing import Callable, List

import numpy as np
from scipy.optimize import curve_fit
from scipy.stats import halfnorm, invgamma
from scipy.stats._distn_infrastructure import rv_frozen  # noqa

__all__ = ["make_invgamma_prior", "roundflat", "create_priors"]


def roundflat(x, a_low=2.0, a_high=8.0, d_low=0.005, d_high=1.2):
    """Return the log probability of the round flat prior.

    The round flat prior is completely uninformative inside the interval bounds
    ``d_low`` and ``d_high`` while smoothly going to -inf for values outside.
    ``a_low`` and ``a_high`` specify how quickly the density falls at the boundaries.

    Args:
        x (float): A parameter value in [0, inf) for which to compute the log
            probability
        a_low (float): Steepness of the prior at the boundary ``d_low``.
        a_high (float): Steepness of the prior at the boundary ``d_high``.
        d_low (float): Lower boundary for which the log probability is -2.
        d_high (float): Upper boundary for which the log probability is -2.

    Returns:
        The log probability for x.
    """
    if x <= 0:
        return -np.inf
    return -2 * ((x / d_low) ** (-2 * a_low) + (x / d_high) ** (2 * a_high))


def make_invgamma_prior(
    lower_bound: float = 0.1, upper_bound: float = 0.5
) -> rv_frozen:
    """Create an inverse gamma distribution prior with 98% density inside the bounds.

    Not all combinations of (lower_bound, upper_bound) are feasible and some of them
    could result in a RuntimeError.

    Parameters
    ----------
    lower_bound : float, default=0.1
        Lower bound at which 1 % of the cumulative density is reached.
    upper_bound : float, default=0.5
        Upper bound at which 99 % of the cumulative density is reached.

    Returns
    -------
    scipy.stats._distn_infrastructure.rv_frozen
        The frozen distribution with shape parameters already set.

    Raises
    ------
    ValueError
        Either if any of the bounds is 0 or negative, or if the upper bound is equal or
        smaller than the lower bound.
    """
    if lower_bound <= 0 or upper_bound <= 0:
        raise ValueError("The bounds cannot be equal to or smaller than 0.")
    if lower_bound >= upper_bound:
        raise ValueError(
            "Lower bound needs to be strictly smaller than the upper " "bound."
        )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        (a_out, scale_out), pcov = curve_fit(
            lambda xdata, a, scale: invgamma.ppf(xdata, a=a, scale=scale),
            [0.01, 0.99],
            [lower_bound, upper_bound],
        )
    return invgamma(a=a_out, scale=scale_out)


def create_priors(
    n_parameters: int,
    signal_scale: float = 4.0,
    lengthscale_lower_bound: float = 0.1,
    lengthscale_upper_bound: float = 0.5,
    noise_scale: float = 0.0006,
) -> List[Callable[[float], float]]:
    """Create a list of priors to be used for the hyperparameters of the tuning process.

    Parameters
    ----------
    n_parameters : int
        Number of parameters to be optimized.
    signal_scale : float
        Prior scale of the signal (standard deviation) which is used to parametrize a
        half-normal distribution.
    lengthscale_lower_bound : float
        Lower bound of the inverse-gamma lengthscale prior. It marks the point at which
        1 % of the cumulative density is reached.
    lengthscale_upper_bound : float
        Upper bound of the inverse-gamma lengthscale prior. It marks the point at which
        99 % of the cumulative density is reached.
    noise_scale : float
        Prior scale of the noise (standard deviation) which is used to parametrize a
        half-normal distribution.

    Returns
    -------
    list of callables
        List of priors in the following order:
         - signal prior
         - lengthscale prior (n_parameters times)
         - noise prior
    """
    if signal_scale <= 0.0:
        raise ValueError(
            f"The signal scale needs to be strictly positive. Got {signal_scale}."
        )
    if noise_scale <= 0.0:
        raise ValueError(
            f"The noise scale needs to be strictly positive. Got {noise_scale}."
        )
    signal_prior = halfnorm(scale=signal_scale)
    lengthscale_prior = make_invgamma_prior(
        lower_bound=lengthscale_lower_bound, upper_bound=lengthscale_upper_bound
    )
    noise_prior = halfnorm(scale=noise_scale)

    priors = [lambda x: signal_prior.logpdf(np.sqrt(np.exp(x))) + x / 2.0 - np.log(2.0)]
    for _ in range(n_parameters):
        priors.append(lambda x: lengthscale_prior.logpdf(np.exp(x)) + x)
    priors.append(
        lambda x: noise_prior.logpdf(np.sqrt(np.exp(x))) + x / 2.0 - np.log(2.0)
    )
    return priors



"""Console script for chess_tuning_tools."""
import json
import logging
import sys
from datetime import datetime

import click
import dill
import numpy as np
from atomicwrites import AtomicWriter
from skopt.utils import create_result

import tune
from tune.db_workers import TuningClient, TuningServer
from tune.io import load_tuning_config, prepare_engines_json, write_engines_json
from tune.local import (
    check_log_for_errors,
    initialize_data,
    initialize_optimizer,
    is_debug_log,
    load_points_to_evaluate,
    parse_experiment_result,
    plot_results,
    print_results,
    run_match,
    setup_logger,
    update_model,
)
from tune.priors import create_priors


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "--verbose", "-v", is_flag=True, default=False, help="Turn on debug output."
)
@click.option("--logfile", default=None, help="Path to where the log is saved to.")
@click.option(
    "--terminate-after", default=0, help="Terminate the client after x minutes."
)
@click.option(
    "--run-only-once",
    default=False,
    is_flag=True,
    help="Terminate the client after one job has been completed or no job can be "
    "found.",
)
@click.option(
    "--skip-benchmark",
    default=False,
    is_flag=True,
    help="Skip calibrating the time control by running a benchmark.",
)
@click.option(
    "--clientconfig", default=None, help="Path to the client configuration file."
)
@click.argument("dbconfig")
def run_client(
    verbose,
    logfile,
    terminate_after,
    run_only_once,
    skip_benchmark,
    clientconfig,
    dbconfig,
):
    """ Run the client to generate games for distributed tuning.

    In order to connect to the database you need to provide a valid DBCONFIG
    json file. It contains the necessary parameters to connect to the database
    where it can fetch jobs and store results.
    """
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        level=log_level,
        filename=logfile,
        format="%(asctime)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    tc = TuningClient(
        dbconfig_path=dbconfig,
        terminate_after=terminate_after,
        clientconfig=clientconfig,
        only_run_once=run_only_once,
        skip_benchmark=skip_benchmark,
    )
    tc.run()


@cli.command()
@click.option(
    "--verbose", "-v", is_flag=True, default=False, help="Turn on debug output."
)
@click.option("--logfile", default=None, help="Path to where the log is saved to.")
@click.argument("command")
@click.argument("experiment_file")
@click.argument("dbconfig")
def run_server(verbose, logfile, command, experiment_file, dbconfig):
    """Run the tuning server for a given EXPERIMENT_FILE (json).

    To connect to the database you also need to provide a DBCONFIG json file.

    \b
    You can choose from these COMMANDs:
     * run: Starts the server.
     * deactivate: Deactivates all active jobs of the given experiment.
     * reactivate: Reactivates all recent jobs for which sample size is not reached yet.
    """
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        level=log_level,
        filename=logfile,
        format="%(asctime)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    tc = TuningServer(experiment_path=experiment_file, dbconfig_path=dbconfig)
    if command == "run":
        tc.run()
    elif command == "deactivate":
        tc.deactivate()
    elif command == "reactivate":
        tc.reactivate()
    else:
        raise ValueError(f"Command {command} is not recognized. Terminating...")


@cli.command()
@click.option(
    "-c",
    "--tuning-config",
    help="json file containing the tuning configuration.",
    required=True,
    type=click.File("r"),
)
@click.option(
    "-a",
    "--acq-function",
    default="mes",
    help="Acquisition function to use for selecting points to try. "
    "Can be one of: {mes, pvrs, ei, ts, vr, lcb, mean, ttei} "
    "Consult the parameter reference and the FAQ for more information.",
    show_default=True,
)
@click.option(
    "--acq-function-samples",
    default=1,
    help="How many GP samples to average the acquisition function over. "
    "More samples will slow down the computation time, but might give more "
    "stable tuning results. Less samples on the other hand cause more exploration "
    "which could help avoid the tuning to get stuck.",
    show_default=True,
)
@click.option(
    "--confidence",
    default=0.90,
    help="Confidence to use for the highest density intervals of the optimum.",
    show_default=True,
)
@click.option(
    "-d",
    "--data-path",
    default="data.npz",
    help="Save the evaluated points to this file.",
    type=click.Path(exists=False),
    show_default=True,
)
@click.option(
    "--gp-burnin",
    default=5,
    type=int,
    help="Number of samples to discard before sampling model parameters. "
    "This is used during tuning and few samples suffice.",
    show_default=True,
)
@click.option(
    "--gp-samples",
    default=300,
    type=int,
    help="Number of model parameters to sample for the model. "
    "This is used during tuning and it should be a multiple of 100.",
    show_default=True,
)
@click.option(
    "--gp-initial-burnin",
    default=100,
    type=int,
    help="Number of samples to discard before starting to sample the initial model "
    "parameters. This is only used when resuming or for the first model.",
    show_default=True,
)
@click.option(
    "--gp-initial-samples",
    default=300,
    type=int,
    help="Number of model parameters to sample for the initial model. "
    "This is only used when resuming or for the first model. "
    "Should be a multiple of 100.",
    show_default=True,
)
@click.option(
    "--gp-signal-prior-scale",
    default=4.0,
    type=click.FloatRange(min=0.0),
    help="Prior scale of the signal (standard deviation) magnitude which is used to"
    "parametrize a half-normal distribution."
    "Needs to be a number strictly greater than 0.0.",
    show_default=True,
)
@click.option(
    "--gp-noise-prior-scale",
    default=0.0006,
    type=click.FloatRange(min=0.0),
    help="Prior scale of the noise (standard deviation) which is used to parametrize a "
    "half-normal distribution."
    "Needs to be a number strictly greater than 0.0.",
    show_default=True,
)
@click.option(
    "--gp-lengthscale-prior-lb",
    default=0.1,
    type=click.FloatRange(min=0.0),
    help="Lower bound for the inverse-gamma lengthscale prior. "
    "It marks the point where the prior reaches 1% of the cumulative density."
    "Needs to be a number strictly greater than 0.0.",
    show_default=True,
)
@click.option(
    "--gp-lengthscale-prior-ub",
    default=0.5,
    type=click.FloatRange(min=0.0),
    help="Upper bound for the inverse-gamma lengthscale prior. "
    "It marks the point where the prior reaches 99% of the cumulative density."
    "Needs to be a number strictly greater than 0.0 and the lower bound.",
    show_default=True,
)
@click.option(
    "-l",
    "--logfile",
    default="log.txt",
    help="Path to log file.",
    type=click.Path(exists=False),
    show_default=True,
)
@click.option(
    "--n-initial-points",
    default=16,
    help="Size of initial dense set of points to try.",
    show_default=True,
)
@click.option(
    "--n-points",
    default=500,
    help="The number of random points to consider as possible next point. "
    "Less points reduce the computation time of the tuner, but reduce "
    "the coverage of the space.",
    show_default=True,
)
@click.option(
    "-p",
    "--evaluate-points",
    default=None,
    type=click.Path(exists=False),
    help="Path to a .csv file (without header row) with points to evaluate. An optional"
    " last column can be used to request a different number of rounds for each "
    "point.",
    show_default=False,
)
@click.option(
    "--plot-every",
    default=1,
    help="Plot the current optimization landscape every n-th iteration. "
    "Set to 0 to turn it off.",
    show_default=True,
)
@click.option(
    "--plot-path",
    default="plots",
    help="Path to the directory to which the tuner will output plots.",
    show_default=True,
)
@click.option(
    "--random-seed",
    default=0,
    help="Number to seed all internally used random generators.",
    show_default=True,
)
@click.option(
    "--result-every",
    default=1,
    help="Output the actual current optimum every n-th iteration."
    "The further you are in the tuning process, the longer this will take to "
    "compute. Consider increasing this number, if you do not need the output "
    "that often. Set to 0 to turn it off.",
    show_default=True,
)
@click.option(
    "--resume/--no-resume",
    default=True,
    help="Let the optimizer resume, if it finds points it can use.",
    show_default=True,
)
@click.option(
    "--fast-resume/--no-fast-resume",
    default=True,
    help="If set, resume the tuning process with the model in the file specified by"
    " the --model-path. "
    "Note, that a full reinitialization will be performed, if the parameter"
    "ranges have been changed.",
    show_default=True,
)
@click.option(
    "--model-path",
    default="model.pkl",
    help="The current optimizer will be saved for fast resuming to this file.",
    type=click.Path(exists=False),
    show_default=True,
)
@click.option("--verbose", "-v", count=True, default=0, help="Turn on debug output.")
@click.option(
    "--warp-inputs/--no-warp-inputs",
    default=True,
    show_default=True,
    help="If True, let the tuner warp the input space to find a better fit to the "
    "optimization landscape.",
)
def local(  # noqa: C901
    tuning_config,
    acq_function="mes",
    acq_function_samples=1,
    confidence=0.9,
    data_path=None,
    evaluate_points=None,
    gp_burnin=5,
    gp_samples=300,
    gp_initial_burnin=100,
    gp_initial_samples=300,
    gp_signal_prior_scale=4.0,
    gp_noise_prior_scale=0.0006,
    gp_lengthscale_prior_lb=0.1,
    gp_lengthscale_prior_ub=0.5,
    logfile="log.txt",
    n_initial_points=16,
    n_points=500,
    plot_every=1,
    plot_path="plots",
    random_seed=0,
    result_every=1,
    resume=True,
    fast_resume=True,
    model_path="model.pkl",
    verbose=0,
    warp_inputs=True,
):
    """Run a local tune.

    Parameters defined in the `tuning_config` file always take precedence.
    """

    json_dict = json.load(tuning_config)
    settings, commands, fixed_params, param_ranges = load_tuning_config(json_dict)
    root_logger = setup_logger(
        verbose=verbose, logfile=settings.get("logfile", logfile)
    )
    # First log the version of chess-tuning-tools:
    root_logger.info(f"chess-tuning-tools version: {tune.__version__}")
    root_logger.debug(f"Got the following tuning settings:\n{json_dict}")

    # Initialize/import data structures:
    if data_path is None:
        data_path = "data.npz"
    try:
        X, y, noise, iteration, optima, performance = initialize_data(
            parameter_ranges=list(param_ranges.values()),
            resume=resume,
            data_path=data_path,
        )
    except ValueError:
        root_logger.error(
            "The number of parameters are not matching the number of "
            "dimensions. Rename the existing data file or ensure that the "
            "parameter ranges are correct."
        )
        sys.exit(1)

    # Initialize Optimizer object and if applicable, resume from existing
    # data/optimizer:
    gp_priors = create_priors(
        n_parameters=len(param_ranges),
        signal_scale=settings.get("gp_signal_prior_scale", gp_signal_prior_scale),
        lengthscale_lower_bound=settings.get(
            "gp_lengthscale_prior_lb", gp_lengthscale_prior_lb
        ),
        lengthscale_upper_bound=settings.get(
            "gp_lengthscale_prior_ub", gp_lengthscale_prior_ub
        ),
        noise_scale=settings.get("gp_noise_prior_scale", gp_noise_prior_scale),
    )
    opt = initialize_optimizer(
        X=X,
        y=y,
        noise=noise,
        parameter_ranges=list(param_ranges.values()),
        random_seed=settings.get("random_seed", random_seed),
        warp_inputs=settings.get("warp_inputs", warp_inputs),
        n_points=settings.get("n_points", n_points),
        n_initial_points=settings.get("n_initial_points", n_initial_points),
        acq_function=settings.get("acq_function", acq_function),
        acq_function_samples=settings.get("acq_function_samples", acq_function_samples),
        resume=resume,
        fast_resume=fast_resume,
        model_path=model_path,
        gp_initial_burnin=settings.get("gp_initial_burnin", gp_initial_burnin),
        gp_initial_samples=settings.get("gp_initial_samples", gp_initial_samples),
        gp_priors=gp_priors,
    )
    extra_points = load_points_to_evaluate(
        space=opt.space, csv_file=evaluate_points, rounds=settings.get("rounds", 10),
    )
    root_logger.debug(
        f"Loaded {len(extra_points)} extra points to evaluate: {extra_points}"
    )

    # Main optimization loop:
    while True:
        root_logger.info("Starting iteration {}".format(iteration))

        # If a model has been fit, print/plot results so far:
        if len(y) > 0 and opt.gp.chain_ is not None:
            result_object = create_result(Xi=X, yi=y, space=opt.space, models=[opt.gp])
            result_every_n = settings.get("result_every", result_every)
            if result_every_n > 0 and iteration % result_every_n == 0:
                try:
                    current_optimum, estimated_elo, estimated_std = print_results(
                        optimizer=opt,
                        result_object=result_object,
                        parameter_names=list(param_ranges.keys()),
                        confidence=settings.get("confidence", confidence),
                    )
                    optima.append(current_optimum)
                    performance.append((int(iteration), estimated_elo, estimated_std))
                except ValueError:
                    pass
            plot_every_n = settings.get("plot_every", plot_every)
            if plot_every_n > 0 and iteration % plot_every_n == 0:
                plot_results(
                    optimizer=opt,
                    result_object=result_object,
                    iterations=np.array(performance)[:, 0],
                    elos=np.array(performance)[:, 1:],
                    optima=np.array(optima),
                    plot_path=settings.get("plot_path", plot_path),
                    parameter_names=list(param_ranges.keys()),
                    confidence=settings.get("confidence", confidence),
                    current_iteration=iteration,
                )

        used_extra_point = False
        # If there are extra points to evaluate, evaluate them first in FIFO order:
        if len(extra_points) > 0:
            point, n_rounds = extra_points.pop(0)
            # Log that we are evaluating the extra point:
            root_logger.info(
                f"Evaluating extra point {dict(zip(param_ranges.keys(), point))} for "
                f"{n_rounds} rounds."
            )
            used_extra_point = True
        else:
            # Ask optimizer for next point:
            point = opt.ask()
            n_rounds = settings.get("rounds", 10)
        match_settings = settings.copy()
        match_settings["rounds"] = n_rounds
        point_dict = dict(zip(param_ranges.keys(), point))
        root_logger.info("Testing {}".format(point_dict))

        # Prepare engines.json file for cutechess-cli:
        engine_json = prepare_engines_json(commands=commands, fixed_params=fixed_params)
        root_logger.debug(f"engines.json is prepared:\n{engine_json}")
        write_engines_json(engine_json, point_dict)

        # Run experiment:
        root_logger.info("Start experiment")
        now = datetime.now()
        out_exp = []
        out_all = []
        for output_line in run_match(**match_settings):
            line = output_line.rstrip()
            is_debug = is_debug_log(line)
            if is_debug and verbose > 2:
                root_logger.debug(line)
            if not is_debug:
                out_exp.append(line)
            out_all.append(line)
        check_log_for_errors(cutechess_output=out_all)
        out_exp = "\n".join(out_exp)
        later = datetime.now()
        difference = (later - now).total_seconds()
        root_logger.info(f"Experiment finished ({difference}s elapsed).")

        # Parse cutechess-cli output and report results (Elo and standard deviation):
        score, error_variance, draw_rate = parse_experiment_result(out_exp, **settings)
        root_logger.info(
            "Got Elo: {} +- {}".format(-score * 100, np.sqrt(error_variance) * 100)
        )
        root_logger.info("Estimated draw rate: {:.2%}".format(draw_rate))

        # Update model with the new data:
        root_logger.info("Updating model")
        update_model(
            optimizer=opt,
            point=point,
            score=score,
            variance=error_variance,
            acq_function_samples=settings.get(
                "acq_function_samples", acq_function_samples
            ),
            gp_burnin=settings.get("gp_burnin", gp_burnin),
            gp_samples=settings.get("gp_samples", gp_samples),
            gp_initial_burnin=settings.get("gp_initial_burnin", gp_initial_burnin),
            gp_initial_samples=settings.get("gp_initial_samples", gp_initial_samples),
        )
        # If we used an extra point, we need to reset n_initial_points of the optimizer:
        if used_extra_point:
            opt._n_initial_points += 1

        # Update data structures and persist to disk:
        X.append(point)
        y.append(score)
        noise.append(error_variance)
        iteration += 1

        with AtomicWriter(data_path, mode="wb", overwrite=True).open() as f:
            np.savez_compressed(
                f,
                np.array(X),
                np.array(y),
                np.array(noise),
                np.array(optima),
                np.array(performance),
                np.array(iteration),
            )
        with AtomicWriter(model_path, mode="wb", overwrite=True).open() as f:
            dill.dump(opt, f)


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



import itertools
from typing import Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.ticker import LogLocator
from scipy.optimize import OptimizeResult
from skopt.plots import _format_scatter_plot_axes
from skopt.space import Space

from tune.utils import confidence_to_mult, expected_ucb, latest_iterations

__all__ = [
    "partial_dependence",
    "plot_objective",
    "plot_objective_1d",
    "plot_optima",
    "plot_performance",
]


def _evenly_sample(dim, n_points):
    """Return `n_points` evenly spaced points from a Dimension.
    Parameters
    ----------
    dim : `Dimension`
        The Dimension to sample from.  Can be categorical; evenly-spaced
        category indices are chosen in order without replacement (result
        may be smaller than `n_points`).
    n_points : int
        The number of points to sample from `dim`.
    Returns
    -------
    xi : np.array
        The sampled points in the Dimension.  For Categorical
        dimensions, returns the index of the value in
        `dim.categories`.
    xi_transformed : np.array
        The transformed values of `xi`, for feeding to a model.
    """
    cats = np.array(getattr(dim, "categories", []), dtype=object)
    if len(cats):  # Sample categoricals while maintaining order
        xi = np.linspace(0, len(cats) - 1, min(len(cats), n_points), dtype=int)
        xi_transformed = dim.transform(cats[xi])
    else:
        bounds = dim.bounds
        # XXX use linspace(*bounds, n_points) after python2 support ends
        if dim.prior == "log-uniform":
            xi = np.logspace(np.log10(bounds[0]), np.log10(bounds[1]), n_points)
        else:
            xi = np.linspace(bounds[0], bounds[1], n_points)
        xi_transformed = dim.transform(xi)
    return xi, xi_transformed


def partial_dependence(
    space, model, i, j=None, sample_points=None, n_samples=250, n_points=40, x_eval=None
):
    """Calculate the partial dependence for dimensions `i` and `j` with
    respect to the objective value, as approximated by `model`.
    The partial dependence plot shows how the value of the dimensions
    `i` and `j` influence the `model` predictions after "averaging out"
    the influence of all other dimensions.
    When `x_eval` is not `None`, the given values are used instead of
    random samples. In this case, `n_samples` will be ignored.
    Parameters
    ----------
    space : `Space`
        The parameter space over which the minimization was performed.
    model
        Surrogate model for the objective function.
    i : int
        The first dimension for which to calculate the partial dependence.
    j : int, default=None
        The second dimension for which to calculate the partial dependence.
        To calculate the 1D partial dependence on `i` alone set `j=None`.
    sample_points : np.array, shape=(n_points, n_dims), default=None
        Only used when `x_eval=None`, i.e in case partial dependence should
        be calculated.
        Randomly sampled and transformed points to use when averaging
        the model function at each of the `n_points` when using partial
        dependence.
    n_samples : int, default=100
        Number of random samples to use for averaging the model function
        at each of the `n_points` when using partial dependence. Only used
        when `sample_points=None` and `x_eval=None`.
    n_points : int, default=40
        Number of points at which to evaluate the partial dependence
        along each dimension `i` and `j`.
    x_eval : list, default=None
        `x_eval` is a list of parameter values or None. In case `x_eval`
        is not None, the parsed dependence will be calculated using these
        values.
        Otherwise, random selected samples will be used.
    Returns
    -------
    For 1D partial dependence:
    xi : np.array
        The points at which the partial dependence was evaluated.
    yi : np.array
        The value of the model at each point `xi`.
    For 2D partial dependence:
    xi : np.array, shape=n_points
        The points at which the partial dependence was evaluated.
    yi : np.array, shape=n_points
        The points at which the partial dependence was evaluated.
    zi : np.array, shape=(n_points, n_points)
        The value of the model at each point `(xi, yi)`.
    For Categorical variables, the `xi` (and `yi` for 2D) returned are
    the indices of the variable in `Dimension.categories`.
    """
    # The idea is to step through one dimension, evaluating the model with
    # that dimension fixed and averaging either over random values or over
    # the given ones in x_val in all other dimensions.
    # (Or step through 2 dimensions when i and j are given.)
    # Categorical dimensions make this interesting, because they are one-
    # hot-encoded, so there is a one-to-many mapping of input dimensions
    # to transformed (model) dimensions.

    # If we haven't parsed an x_eval list we use random sampled values instead
    if x_eval is None and sample_points is None:
        sample_points = space.transform(space.rvs(n_samples=n_samples))
    elif sample_points is None:
        sample_points = space.transform([x_eval])

    # dim_locs[i] is the (column index of the) start of dim i in
    # sample_points.
    # This is usefull when we are using one hot encoding, i.e using
    # categorical values
    dim_locs = np.cumsum([0] + [d.transformed_size for d in space.dimensions])

    if j is None:
        # We sample evenly instead of randomly. This is necessary when using
        # categorical values
        xi, xi_transformed = _evenly_sample(space.dimensions[i], n_points)
        yi = []
        for x_ in xi_transformed:
            rvs_ = np.array(sample_points)  # copy
            # We replace the values in the dimension that we want to keep
            # fixed
            rvs_[:, dim_locs[i] : dim_locs[i + 1]] = x_
            # In case of `x_eval=None` rvs conists of random samples.
            # Calculating the mean of these samples is how partial dependence
            # is implemented.
            yi.append(np.mean(model.predict(rvs_)))

        return xi, yi

    else:
        xi, xi_transformed = _evenly_sample(space.dimensions[j], n_points)
        yi, yi_transformed = _evenly_sample(space.dimensions[i], n_points)

        zi = []
        for x_ in xi_transformed:
            row = []
            for y_ in yi_transformed:
                rvs_ = np.array(sample_points)  # copy
                rvs_[:, dim_locs[j] : dim_locs[j + 1]] = x_
                rvs_[:, dim_locs[i] : dim_locs[i + 1]] = y_
                row.append(np.mean(model.predict(rvs_)))
            zi.append(row)

        return xi, yi, np.array(zi).T


def plot_objective_1d(
    result: OptimizeResult,
    parameter_name: Optional[str] = None,
    n_points: int = 500,
    n_random_restarts: int = 100,
    confidence: float = 0.9,
    figsize: Tuple[float, float] = (10, 6),
    fig: Optional[Figure] = None,
    ax: Optional[Axes] = None,
    colors: Optional[Sequence[Union[tuple, str]]] = None,
) -> Tuple[Figure, Axes]:
    """Plot the 1D objective function.

    Parameters
    ----------
    result : OptimizeResult
        The current optimization result.
    parameter_name : str, optional
        The name of the parameter to plot. If None, no x-axis label is shown.
    n_points : int (default=500)
        The number of points to use for prediction of the Gaussian process.
    n_random_restarts : int (default=100)
        The number of random restarts to employ to find the optima.
    confidence : float (default=0.9)
        The confidence interval to plot around the mean prediction.
    figsize : tuple (default=(10, 6))
        The size of the figure.
    fig : Figure, optional
        The figure to use. If None, a new figure is created.
    ax : Axes, optional
        The axes to use. If None, new axes are created.
    colors : Sequence of colors, optional
        The colors to use for different elements in the plot.
        Can be tuples or strings.

    Returns
    -------
    fig : Figure
        The figure.
    ax : Axes
        The axes.

    """
    if colors is None:
        colors = plt.cm.get_cmap("Set3").colors

    if fig is None:
        plt.style.use("dark_background")
        gs_kw = dict(width_ratios=(1,), height_ratios=[5, 1], hspace=0.05)
        fig, ax = plt.subplots(figsize=figsize, nrows=2, gridspec_kw=gs_kw, sharex=True)
        for a in ax:
            a.set_facecolor("#36393f")
        fig.patch.set_facecolor("#36393f")
    gp = result.models[-1]

    # Compute the optima of the objective function:
    failures = 0
    while True:
        try:
            with gp.noise_set_to_zero():
                min_x = expected_ucb(
                    result, alpha=0.0, n_random_starts=n_random_restarts
                )[0]
                min_ucb = expected_ucb(result, n_random_starts=n_random_restarts)[0]
        except ValueError:
            failures += 1
            if failures == 10:
                break
            continue
        else:
            break

    # Regardless of the range of the parameter to be plotted, the model always operates
    # in [0, 1]:
    x_gp = np.linspace(0, 1, num=n_points)
    x_orig = np.array(result.space.inverse_transform(x_gp[:, None])).flatten()
    with gp.noise_set_to_zero():
        y, y_err = gp.predict(x_gp[:, None], return_std=True)
    y = -y * 100
    y_err = y_err * 100
    confidence_mult = confidence_to_mult(confidence)

    (mean_plot,) = ax[0].plot(x_orig, y, zorder=4, color=colors[0])
    err_plot = ax[0].fill_between(
        x_orig,
        y - y_err * confidence_mult,
        y + y_err * confidence_mult,
        alpha=0.3,
        zorder=0,
        color=colors[0],
    )
    opt_plot = ax[0].axvline(x=min_x, zorder=3, color=colors[3])
    pess_plot = ax[0].axvline(x=min_ucb, zorder=2, color=colors[5])
    if parameter_name is not None:
        ax[1].set_xlabel(parameter_name)
    dim = result.space.dimensions[0]
    ax[0].set_xlim(dim.low, dim.high)
    match_plot = ax[1].scatter(
        x=result.x_iters,
        y=-result.func_vals * 100,
        zorder=1,
        marker=".",
        s=0.6,
        color=colors[0],
    )
    ax[0].set_ylabel("Elo")
    ax[1].set_ylabel("Elo")
    fig.legend(
        (mean_plot, err_plot, opt_plot, pess_plot, match_plot),
        ("Mean", f"{confidence:.0%} CI", "Optimum", "Conservative Optimum", "Matches"),
        loc="lower center",
        ncol=5,
        bbox_to_anchor=(0.5, -0.03),
        frameon=False,
    )

    return fig, ax


def plot_objective(
    result,
    levels=20,
    n_points=200,
    n_samples=30,
    size=3,
    zscale="linear",
    dimensions=None,
    n_random_restarts=100,
    alpha=0.25,
    margin=0.65,
    colors=None,
    fig=None,
    ax=None,
):
    """Pairwise partial dependence plot of the objective function.
    The diagonal shows the partial dependence for dimension `i` with
    respect to the objective function. The off-diagonal shows the
    partial dependence for dimensions `i` and `j` with
    respect to the objective function. The objective function is
    approximated by `result.model.`
    Pairwise scatter plots of the points at which the objective
    function was directly evaluated are shown on the off-diagonal.
    A red point indicates the found minimum.
    Note: search spaces that contain `Categorical` dimensions are
          currently not supported by this function.
    Parameters
    ----------
    * `result` [`OptimizeResult`]
        The result for which to create the scatter plot matrix.
    * `levels` [int, default=10]
        Number of levels to draw on the contour plot, passed directly
        to `plt.contour()`.
    * `n_points` [int, default=40]
        Number of points at which to evaluate the partial dependence
        along each dimension.
    * `n_samples` [int, default=250]
        Number of random samples to use for averaging the model function
        at each of the `n_points`.
    * `size` [float, default=2]
        Height (in inches) of each facet.
    * `zscale` [str, default='linear']
        Scale to use for the z axis of the contour plots. Either 'linear'
        or 'log'.
    * `dimensions` [list of str, default=None] Labels of the dimension
        variables. `None` defaults to `space.dimensions[i].name`, or
        if also `None` to `['X_0', 'X_1', ..]`.
    * `n_random_restarts` [int, default=100]
        Number of restarts to try to find the global optimum.
    * `alpha` [float, default=0.25]
        Transparency of the sampled points.
    * `margin` [float, default=0.65]
        Margin in inches around the plot.
    * `colors` [list of tuples, default=None]
        Colors to use for the optima.
    * `fig` [Matplotlib figure, default=None]
        Figure to use for plotting. If None, it will create one.
    * `ax` [k x k axes, default=None]
        Axes on which to plot the marginals. If None, it will create appropriate
        axes.
    Returns
    -------
    * `ax`: [`Axes`]:
        The matplotlib axes.
    """
    if colors is None:
        colors = plt.cm.get_cmap("Set3").colors
    space = result.space
    samples = np.asarray(result.x_iters)
    rvs_transformed = space.transform(space.rvs(n_samples=n_samples))

    if zscale == "log":
        locator = LogLocator()
    elif zscale == "linear":
        locator = None
    else:
        raise ValueError(
            "Valid values for zscale are 'linear' and 'log'," " not '%s'." % zscale
        )
    if fig is None:
        fig, ax = plt.subplots(
            space.n_dims,
            space.n_dims,
            figsize=(size * space.n_dims, size * space.n_dims),
        )
    width, height = fig.get_size_inches()

    fig.subplots_adjust(
        left=margin / width,
        right=1 - margin / width,
        bottom=margin / height,
        top=1 - margin / height,
        hspace=0.1,
        wspace=0.1,
    )
    failures = 0
    while True:
        try:
            with result.models[-1].noise_set_to_zero():
                min_x = expected_ucb(
                    result, alpha=0.0, n_random_starts=n_random_restarts
                )[0]
                min_ucb = expected_ucb(result, n_random_starts=n_random_restarts)[0]
        except ValueError:
            failures += 1
            if failures == 10:
                break
            continue
        else:
            break

    for i in range(space.n_dims):
        for j in range(space.n_dims):
            if i == j:
                xi, yi = partial_dependence(
                    space,
                    result.models[-1],
                    i,
                    j=None,
                    sample_points=rvs_transformed,
                    n_points=n_points,
                )
                yi_min, yi_max = np.min(yi), np.max(yi)
                ax[i, i].plot(xi, yi, color=colors[1])
                if failures != 10:
                    ax[i, i].axvline(min_x[i], linestyle="--", color=colors[3], lw=1)
                    ax[i, i].axvline(min_ucb[i], linestyle="--", color=colors[5], lw=1)
                    ax[i, i].text(
                        min_x[i],
                        yi_min + 0.9 * (yi_max - yi_min),
                        f"{np.around(min_x[i], 4)}",
                        color=colors[3],
                    )
                    ax[i, i].text(
                        min_ucb[i],
                        yi_min + 0.7 * (yi_max - yi_min),
                        f"{np.around(min_ucb[i], 4)}",
                        color=colors[5],
                    )

            # lower triangle
            elif i > j:
                xi, yi, zi = partial_dependence(
                    space, result.models[-1], i, j, rvs_transformed, n_points
                )
                ax[i, j].contourf(xi, yi, zi, levels, locator=locator, cmap="viridis_r")
                ax[i, j].scatter(
                    samples[:, j], samples[:, i], c="k", s=10, lw=0.0, alpha=alpha
                )
                if failures != 10:
                    ax[i, j].scatter(min_x[j], min_x[i], c=["r"], s=20, lw=0.0)
                    ax[i, j].scatter(
                        min_ucb[j], min_ucb[i], c=["xkcd:orange"], s=20, lw=0.0
                    )
    # Get all dimensions.
    plot_dims = []
    for row in range(space.n_dims):
        if space.dimensions[row].is_constant:
            continue
        plot_dims.append((row, space.dimensions[row]))
    return _format_scatter_plot_axes(
        ax,
        space,
        ylabel="Partial dependence",
        plot_dims=plot_dims,
        dim_labels=dimensions,
    )


def plot_optima(
    iterations: np.ndarray,
    optima: np.ndarray,
    space: Optional[Space] = None,
    parameter_names: Optional[Sequence[str]] = None,
    plot_width: float = 8,
    aspect_ratio: float = 0.4,
    fig: Optional[Figure] = None,
    ax: Optional[Union[Axes, np.ndarray]] = None,
    colors: Optional[Sequence[Union[tuple, str]]] = None,
) -> Tuple[Figure, np.ndarray]:
    """Plot the optima found by the tuning algorithm.

    Parameters
    ----------
    iterations : np.ndarray
        The iterations at which the optima were found.
    optima : np.ndarray
        The optima found recorded at the given iterations.
    space : Space, optional
        The optimization space for the parameters. If provided, it will be used to
        scale the y-axes and to apply log-scaling, if the parameter is optimized on
        a log-scale.
    parameter_names : Sequence[str], optional
        The names of the parameters. If not provided, no y-axis labels will be shown.
    plot_width : int, optional
        The width of each plot in inches. The total width of the plot will be larger
        depending on the number of parameters and how they are arranged.
    aspect_ratio : float, optional
        The aspect ratio of the subplots. The default is 0.4, which means that the
        height of each subplot will be 40% of the width.
    fig : Figure, optional
        The figure to plot on. If not provided, a new figure in the style of
        chess-tuning-tools will be created.
    ax : np.ndarray or Axes, optional
        The axes to plot on. If not provided, new axes will be created.
        If provided, the axes will be filled. Thus, the number of axes should be at
        least as large as the number of parameters.
    colors : Sequence[Union[tuple, str]], optional
        The colors to use for the plots. If not provided, the color scheme 'Set3' of
        matplotlib will be used.

    Returns
    -------
    Figure
        The figure containing the plots.
    np.ndarray
        A two-dimensional array containing the axes.

    Raises
    ------
    ValueError
        - if the number of parameters does not match the number of parameter names
        - if the number of axes is smaller than the number of parameters
        - if the number of iterations is not matching the number of optima
        - if a fig, but no ax is passed
    """
    if optima.shape[0] != len(iterations):
        raise ValueError("Iteration array does not match optima array.")
    iterations, optima = latest_iterations(iterations, optima)
    n_points, n_parameters = optima.shape
    if parameter_names is not None and len(parameter_names) != n_parameters:
        raise ValueError(
            "Number of parameter names does not match the number of parameters."
        )
    if colors is None:
        colors = plt.cm.get_cmap("Set3").colors
    n_colors = len(colors)
    if fig is None:
        plt.style.use("dark_background")
        n_cols = int(np.floor(np.sqrt(n_parameters)))
        n_rows = int(np.ceil(n_parameters / n_cols))
        figsize = (n_cols * plot_width, aspect_ratio * plot_width * n_rows)
        fig, ax = plt.subplots(
            figsize=figsize, nrows=n_rows, ncols=n_cols, sharex=True,
        )

        margin_left = 1.0
        margin_right = 0.1
        margin_bottom = 0.5
        margin_top = 0.4
        wspace = 1
        hspace = 0.3
        plt.subplots_adjust(
            left=margin_left / figsize[0],
            right=1 - margin_right / figsize[0],
            bottom=margin_bottom / figsize[1],
            top=1 - margin_top / figsize[1],
            hspace=n_rows * hspace / figsize[1],
            wspace=n_cols * wspace / figsize[0],
        )
        ax = np.atleast_2d(ax).reshape(n_rows, n_cols)
        for a in ax.reshape(-1):
            a.set_facecolor("#36393f")
            a.grid(which="major", color="#ffffff", alpha=0.1)
        fig.patch.set_facecolor("#36393f")
        fig.suptitle(
            "Predicted best parameters over time",
            y=1 - 0.5 * margin_top / figsize[1],
            va="center",
        )
    else:
        if ax is None:
            raise ValueError("Axes must be specified if a figure is provided.")
        if not hasattr(ax, "__len__"):
            n_rows = n_cols = 1
        elif ax.ndim == 1:
            n_rows = len(ax)
            n_cols = 1
        else:
            n_rows, n_cols = ax.shape
        if n_rows * n_cols < n_parameters:
            raise ValueError("Not enough axes to plot all parameters.")
        ax = np.atleast_2d(ax).reshape(n_rows, n_cols)

    for i, (j, k) in enumerate(itertools.product(range(n_rows), range(n_cols))):
        a = ax[j, k]
        if i >= n_parameters:
            fig.delaxes(a)
            continue
        # If the axis is the last one in the current column, then set the xlabel:
        if (j + 1) * n_cols + k + 1 > n_parameters:
            a.set_xlabel("Iteration")
            # Since hspace=0, we have to fix the xaxis label and tick labels here:
            a.xaxis.set_label_coords(0.5, -0.12)
            a.xaxis.set_tick_params(labelbottom=True)

        a.plot(
            iterations,
            optima[:, i],
            color=colors[i % n_colors],
            zorder=10,
            linewidth=1.3,
        )
        a.axhline(
            y=optima[-1, i],
            color=colors[i % n_colors],
            zorder=9,
            linewidth=0.5,
            ls="--",
            alpha=0.6,
        )
        # If the user supplied an optimization space, we can use that information to
        # scale the yaxis and apply log-scaling, where necessary:
        s = f"{optima[-1, i]:.2f}"
        if space is not None:
            dim = space.dimensions[i]
            a.set_ylim(*dim.bounds)
            if dim.prior == "log-uniform":
                a.set_yscale("log")
                s = np.format_float_scientific(optima[-1, i], precision=2)

        # Label the horizontal line of the current optimal value:
        # First convert the y-value to normalized axes coordinates:
        point = a.get_xlim()[0], optima[-1, i]
        transformed_point = a.transAxes.inverted().transform(
            a.transData.transform(point)
        )
        a.text(
            x=transformed_point[0] + 0.01,
            y=transformed_point[1] - 0.02,
            s=s,
            bbox=dict(
                facecolor="#36393f",
                edgecolor="None",
                alpha=0.5,
                boxstyle="square,pad=0.1",
            ),
            transform=a.transAxes,
            horizontalalignment="left",
            verticalalignment="top",
            color=colors[i % n_colors],
            zorder=11,
        )

        if parameter_names is not None:
            a.set_ylabel(parameter_names[i])
    return fig, ax


def plot_performance(
    performance: np.ndarray,
    confidence: float = 0.9,
    plot_width: float = 8,
    aspect_ratio: float = 0.7,
    fig: Optional[Figure] = None,
    ax: Optional[Axes] = None,
    colors: Optional[Sequence[Union[tuple, str]]] = None,
) -> Tuple[Figure, np.ndarray]:
    """Plot the estimated Elo of the Optima predicted by the tuning algorithm.

    Parameters
    ----------
    performance : np.ndarray, shape=(n_iterations, 3)
        Array containing the iteration numbers, the estimated Elo of the predicted
        optimum, and the estimated standard error of the estimated Elo.
    confidence : float, optional (default=0.9)
        The confidence interval to plot around the estimated Elo.
    plot_width : int, optional (default=8)
        The width of each plot in inches. The total width of the plot will be larger
        depending on the number of parameters and how they are arranged.
    aspect_ratio : float, optional (default=0.7)
        The aspect ratio of the subplots. The default is 0.4, which means that the
        height of each subplot will be 40% of the width.
    fig : Figure, optional
        The figure to plot on. If not provided, a new figure in the style of
        chess-tuning-tools will be created.
    ax : np.ndarray or Axes, optional
        The axes to plot on. If not provided, new axes will be created.
        If provided, the axes will be filled. Thus, the number of axes should be at
        least as large as the number of parameters.
    colors : Sequence[Union[tuple, str]], optional
        The colors to use for the plots. If not provided, the color scheme 'Set3' of
        matplotlib will be used.

    Returns
    -------
    Figure
        The figure containing the plots.
    np.ndarray
        A two-dimensional array containing the axes.

    Raises
    ------
    ValueError
        - if the number of parameters does not match the number of parameter names
        - if the number of axes is smaller than the number of parameters
        - if the number of iterations is not matching the number of optima
        - if a fig, but no ax is passed
    """
    iterations, elo, elo_std = latest_iterations(*performance.T)
    if colors is None:
        colors = plt.cm.get_cmap("Set3").colors
    if fig is None:
        plt.style.use("dark_background")
        figsize = (plot_width, aspect_ratio * plot_width)
        fig, ax = plt.subplots(figsize=figsize)

        margin_left = 0.8
        margin_right = 0.1
        margin_bottom = 0.7
        margin_top = 0.3
        plt.subplots_adjust(
            left=margin_left / figsize[0],
            right=1 - margin_right / figsize[0],
            bottom=margin_bottom / figsize[1],
            top=1 - margin_top / figsize[1],
        )
        ax.set_facecolor("#36393f")
        ax.grid(which="major", color="#ffffff", alpha=0.1)
        fig.patch.set_facecolor("#36393f")
        ax.set_title("Elo of the predicted best parameters over time")
    elif ax is None:
        raise ValueError("Axes must be specified if a figure is provided.")

    ax.plot(
        iterations,
        elo,
        color=colors[0],
        zorder=10,
        linewidth=1.3,
        label="Predicted Elo",
    )
    confidence_mult = confidence_to_mult(confidence)
    ax.fill_between(
        iterations,
        elo - confidence_mult * elo_std,
        elo + confidence_mult * elo_std,
        color=colors[0],
        linewidth=0,
        zorder=9,
        alpha=0.25,
        label=f"{confidence:.0%} confidence interval",
    )
    ax.axhline(
        y=elo[-1],
        linestyle="--",
        zorder=8,
        color=colors[0],
        label="Last prediction",
        linewidth=1,
        alpha=0.3,
    )
    ax.legend(loc="upper center", frameon=False, bbox_to_anchor=(0.5, -0.08), ncol=3)
    ax.set_xlabel("Iteration")
    ax.set_ylabel("Elo")
    ax.set_xlim(min(iterations), max(iterations))

    return fig, ax



import logging
import pathlib
import re
import subprocess
import sys
import time
from datetime import datetime
from logging import Logger
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    TextIO,
    Tuple,
    Union,
)

import dill
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from bask import Optimizer
from matplotlib.transforms import Bbox
from numpy.random import RandomState
from scipy.optimize import OptimizeResult
from scipy.stats import dirichlet
from skopt.space import Categorical, Dimension, Integer, Real, Space
from skopt.utils import normalize_dimensions

from tune.plots import plot_objective, plot_objective_1d, plot_optima, plot_performance
from tune.summary import confidence_intervals
from tune.utils import TimeControl, confidence_to_mult, expected_ucb

__all__ = [
    "counts_to_penta",
    "initialize_optimizer",
    "run_match",
    "is_debug_log",
    "check_log_for_errors",
    "parse_experiment_result",
    "print_results",
    "plot_results",
    "reduce_ranges",
    "update_model",
    "elo_to_prob",
    "prob_to_elo",
    "setup_logger",
]

LOGGER = "ChessTuner"


def elo_to_prob(elo: np.ndarray, k: float = 4.0) -> np.ndarray:
    """Convert an Elo score (logit space) to a probability.

    Parameters
    ----------
    elo : float
        A real-valued Elo score.
    k : float, optional (default=4.0)
        Scale of the logistic distribution.

    Returns
    -------
    float
        Win probability

    Raises
    ------
    ValueError
        if k <= 0

    """
    if k <= 0:
        raise ValueError("k must be positive")
    return np.atleast_1d(1 / (1 + np.power(10, -elo / k)))


def prob_to_elo(p: np.ndarray, k: float = 4.0) -> np.ndarray:
    """Convert a win probability to an Elo score (logit space).

    Parameters
    ----------
    p : float
        The win probability of the player.
    k : float, optional (default=4.0)
        Scale of the logistic distribution.

    Returns
    -------
    float
        Elo score of the player

    Raises
    ------
    ValueError
        if k <= 0

    """
    if k <= 0:
        raise ValueError("k must be positive")
    return np.atleast_1d(k * np.log10(-p / (p - 1)))


def counts_to_penta(
    counts: np.ndarray,
    prior_counts: Optional[Iterable[float]] = None,
    n_dirichlet_samples: int = 1000000,
    score_scale: float = 4.0,
    random_state: Union[int, RandomState, None] = None,
    **kwargs,
) -> Tuple[float, float]:
    """Compute mean Elo score and variance of the pentanomial model for a count array.

    Parameters
    ----------
    counts : np.ndarray
        Array of counts for WW, WD, WL/DD, LD and LL
    prior_counts : np.ndarray or None, default=None
        Pseudo counts to use for WW, WD, WL/DD, LD and LL in the
        pentanomial model.
    n_dirichlet_samples : int, default = 1 000 000
        Number of samples to draw from the Dirichlet distribution in order to
        estimate the standard error of the score.
    score_scale : float, optional (default=4.0)
        Scale of the logistic distribution used to calculate the score. Has to be a
        positive real number
    random_state : int, RandomState instance or None, optional (default: None)
        The generator used to initialize the centers. If int, random_state is
        the seed used by the random number generator; If RandomState instance,
        random_state is the random number generator; If None, the random number
        generator is the RandomState instance used by `np.random`.
    kwargs : dict
        Additional keyword arguments
    Returns
    -------
    tuple (float, float)
        Mean Elo score and corresponding variance
    """
    if prior_counts is None:
        prior_counts = np.array([0.14, 0.19, 0.34, 0.19, 0.14]) * 2.5
    else:
        prior_counts = np.array(prior_counts)
        if len(prior_counts) != 5:
            raise ValueError("Argument prior_counts should contain 5 elements.")
    dist = dirichlet(alpha=counts + prior_counts)
    scores = [0.0, 0.25, 0.5, 0.75, 1.0]
    score = float(prob_to_elo(dist.mean().dot(scores), k=score_scale))
    error = prob_to_elo(
        dist.rvs(n_dirichlet_samples, random_state=random_state).dot(scores),
        k=score_scale,
    ).var()
    return score, error


def setup_logger(verbose: int = 0, logfile: str = "log.txt") -> Logger:
    """Setup logger with correct verbosity and file handler.

    Parameters
    ----------
    verbose : int
        Verbosity level. If verbose = 0, use INFO level, otherwise DEBUG.
    logfile : str
        Desired path to the logfile.

    Returns
    -------
    Logger
        Logger to be used for logging.
    """
    log_level = logging.DEBUG if verbose > 0 else logging.INFO
    log_format = logging.Formatter("%(asctime)s %(levelname)-8s %(message)s")
    logger = logging.getLogger(LOGGER)
    logger.setLevel(log_level)
    logger.propagate = False

    file_logger = logging.FileHandler(logfile)
    file_logger.setFormatter(log_format)
    logger.addHandler(file_logger)
    console_logger = logging.StreamHandler(sys.stdout)
    console_logger.setFormatter(log_format)
    logger.addHandler(console_logger)
    return logger


def load_points_to_evaluate(
    space: Space, csv_file: Optional[TextIO] = None, rounds: int = 10
) -> List[Tuple[List, int]]:
    """Load extra points to evaluate from a csv file.

    Parameters
    ----------
    space : Space
        Optimization space containing the parameters to tune.
    csv_file : TextIO, optional (default=None)
        Comma-separated text file containing the points to evaluate. If an extra column
        is present, it will be used as the number of rounds to play.
    rounds : int, optional (default=10)
        Number of rounds to play for each point. Will be used if no extra column is
        present.

    Returns
    -------
    List[Tuple[List, int]]
        List of points to evaluate and number of rounds to play each point.

    Raises
    ------
    ValueError
        If the values in the csv file are out of the optimization bounds or if the
        number of columns is not equal to the number of parameters (+1 in the case the
        sample size is provided).
    """

    if csv_file is None:
        # No file given, do not load any points:
        return []

    # Open csv file using pandas:
    df = pd.read_csv(csv_file, header=None)

    # First check if number of columns is correct:
    n_dim = len(space.dimensions)
    if len(df.columns) not in (n_dim, n_dim + 1):
        raise ValueError(
            f"Number of columns in csv file ({len(df.columns)}) does not match number"
            f" of dimensions ({n_dim})."
        )

    # Check if the given points are within the lower and upper bounds of the space:
    for i, dim in enumerate(space.dimensions):
        if not (df.iloc[:, i].between(dim.low, dim.high).all()):
            raise ValueError(
                "Some points in the csv file are outside of the specified bounds."
                " Please check the csv file and your configuration."
            )

    # If there is an extra column, extract it as the number of rounds for each point:
    if len(df.columns) == n_dim + 1:
        rounds_column = df.iloc[:, n_dim].values
        df = df.iloc[:, :n_dim]
    else:
        rounds_column = np.full(len(df), rounds)

    # All points are within the bounds, add them to the list of points to evaluate:
    return [(x, r) for x, r in zip(df.values.tolist(), rounds_column)]


def reduce_ranges(
    X: Sequence[list], y: Sequence[float], noise: Sequence[float], space: Space
) -> Tuple[bool, List[list], List[float], List[float]]:
    """Return all data points consistent with the new restricted space.

    Parameters
    ----------
    X : Sequence of lists
        Contains n_points many lists, each representing one configuration.
    y : Sequence of floats
        Contains n_points many scores, one for each configuration.
    noise : Sequence of floats
        Contains n_points many variances, one for each score.
    space : skopt.space.Space
        Space object specifying the new optimization space.

    Returns
    -------
    Tuple (bool, list, list, list)
        Returns a boolean indicating if a reduction of the dataset was needed and the
        corresponding new X, y and noise lists.
    """
    X_new = []
    y_new = []
    noise_new = []
    reduction_needed = False
    for row, yval, nval in zip(X, y, noise):
        include_row = True
        for dim, value in zip(space.dimensions, row):
            if isinstance(dim, Integer) or isinstance(dim, Real):
                lb, ub = dim.bounds
                if value < lb or value > ub:
                    include_row = False
            elif isinstance(dim, Categorical):
                if value not in dim.bounds:
                    include_row = False
            else:
                raise ValueError(f"Parameter type {type(dim)} unknown.")
        if include_row:
            X_new.append(row)
            y_new.append(yval)
            noise_new.append(nval)
        else:
            reduction_needed = True
    return reduction_needed, X_new, y_new, noise_new


def initialize_data(
    parameter_ranges: Sequence[Union[Sequence, Dimension]],
    data_path: Optional[str] = None,
    resume: bool = True,
) -> Tuple[list, list, list, int, list, list]:
    """Initialize data structures needed for tuning. Either empty or resumed from disk.

    Parameters
    ----------
    parameter_ranges : Sequence of Dimension objects or tuples
        Parameter range specifications as expected by scikit-optimize.
    data_path : str or None, default=None
        Path to the file containing the data structures used for resuming.
        If None, no resuming will be performed.
    resume : bool, default=True
        If True, fill the data structures with the the data from the given data_path.
        Otherwise return empty data structures.

    Returns
    -------
    tuple consisting of list, list, list and int
        Returns the initialized data structures X, y, noise and iteration number.

    Raises
    ------
    ValueError
        If the number of specified parameters is not matching the existing number of
        parameters in the data.
    """
    logger = logging.getLogger()
    X = []
    y = []
    noise = []
    optima = []
    performance = []
    iteration = 0
    if data_path is not None and resume:
        space = normalize_dimensions(parameter_ranges)
        path = pathlib.Path(data_path)
        if path.exists():
            with np.load(path) as importa:
                X = importa["arr_0"].tolist()
                y = importa["arr_1"].tolist()
                noise = importa["arr_2"].tolist()
                if "arr_3" in importa:
                    optima = importa["arr_3"].tolist()
                if "arr_4" in importa:
                    performance = importa["arr_4"].tolist()
                if "arr_5" in importa:
                    iteration = importa["arr_5"]
                else:
                    iteration = len(X)
            if len(X[0]) != space.n_dims:
                raise ValueError(
                    f"Number of parameters ({len(X[0])}) are not matching "
                    f"the number of dimensions ({space.n_dims})."
                )
            reduction_needed, X_reduced, y_reduced, noise_reduced = reduce_ranges(
                X, y, noise, space
            )
            if reduction_needed:
                backup_path = path.parent / (
                    path.stem + f"_backup_{int(time.time())}" + path.suffix
                )
                logger.warning(
                    f"The parameter ranges are smaller than the existing data. "
                    f"Some points will have to be discarded. "
                    f"The original {len(X)} data points will be saved to "
                    f"{backup_path}"
                )
                np.savez_compressed(
                    backup_path, np.array(X), np.array(y), np.array(noise)
                )
                X = X_reduced
                y = y_reduced
                noise = noise_reduced
    return X, y, noise, iteration, optima, performance


def setup_random_state(seed: int) -> np.random.RandomState:
    """Return a seeded RandomState object.

    Parameters
    ----------
    seed : int
        Random seed to be used to seed the RandomState.

    Returns
    -------
    numpy.random.RandomState
        RandomState to be used to generate random numbers.
    """
    ss = np.random.SeedSequence(seed)
    return np.random.RandomState(np.random.MT19937(ss.spawn(1)[0]))


def initialize_optimizer(
    X: Sequence[list],
    y: Sequence[float],
    noise: Sequence[float],
    parameter_ranges: Sequence[Union[Sequence, Dimension]],
    random_seed: int = 0,
    warp_inputs: bool = True,
    n_points: int = 500,
    n_initial_points: int = 16,
    acq_function: str = "mes",
    acq_function_samples: int = 1,
    resume: bool = True,
    fast_resume: bool = True,
    model_path: Optional[str] = None,
    gp_initial_burnin: int = 100,
    gp_initial_samples: int = 300,
    gp_priors: Optional[List[Callable[[float], float]]] = None,
) -> Optimizer:
    """Create an Optimizer object and if needed resume and/or reinitialize.

    Parameters
    ----------
    X : Sequence of lists
        Contains n_points many lists, each representing one configuration.
    y : Sequence of floats
        Contains n_points many scores, one for each configuration.
    noise : Sequence of floats
        Contains n_points many variances, one for each score.
    parameter_ranges : Sequence of Dimension objects or tuples
        Parameter range specifications as expected by scikit-optimize.
    random_seed : int, default=0
        Random seed for the optimizer.
    warp_inputs : bool, default=True
        If True, the optimizer will internally warp the input space for a better model
        fit. Can negatively impact running time and required burnin samples.
    n_points : int, default=500
        Number of points to evaluate the acquisition function on.
    n_initial_points : int, default=16
        Number of points to pick quasi-randomly to initialize the the model, before
        using the acquisition function.
    acq_function : str, default="mes"
        Acquisition function to use.
    acq_function_samples : int, default=1
        Number of hyperposterior samples to average the acquisition function over.
    resume : bool, default=True
        If True, resume optimization from existing data. If False, start with a
        completely fresh optimizer.
    fast_resume : bool, default=True
        If True, restore the optimizer from disk, avoiding costly reinitialization.
        If False, reinitialize the optimizer from the existing data.
    model_path : str or None, default=None
        Path to the file containing the existing optimizer to be used for fast resume
        functionality.
    gp_initial_burnin : int, default=100
        Number of burnin samples to use for reinitialization.
    gp_initial_samples : int, default=300
        Number of samples to use for reinitialization.
    gp_priors : list of callables, default=None
        List of priors to be used for the kernel hyperparameters. Specified in the
        following order:
        - signal magnitude prior
        - lengthscale prior (x number of parameters)
        - noise magnitude prior

    Returns
    -------
    bask.Optimizer
        Optimizer object to be used in the main tuning loop.
    """
    logger = logging.getLogger(LOGGER)
    # Create random generator:
    random_state = setup_random_state(random_seed)

    gp_kwargs = dict(
        # TODO: Due to a bug in scikit-learn 0.23.2, we set normalize_y=False:
        normalize_y=True,
        warp_inputs=warp_inputs,
    )
    opt = Optimizer(
        dimensions=parameter_ranges,
        n_points=n_points,
        n_initial_points=n_initial_points,
        # gp_kernel=kernel,  # TODO: Let user pass in different kernels
        gp_kwargs=gp_kwargs,
        gp_priors=gp_priors,
        acq_func=acq_function,
        acq_func_kwargs=dict(alpha=1.96, n_thompson=500),
        random_state=random_state,
    )

    if not resume:
        return opt

    reinitialize = True
    if model_path is not None and fast_resume:
        path = pathlib.Path(model_path)
        if path.exists():
            with open(model_path, mode="rb") as model_file:
                old_opt = dill.load(model_file)
                logger.info(f"Resuming from existing optimizer in {model_path}.")
            if opt.space == old_opt.space:
                old_opt.acq_func = opt.acq_func
                old_opt.acq_func_kwargs = opt.acq_func_kwargs
                opt = old_opt
                reinitialize = False
            else:
                logger.info(
                    "Parameter ranges have been changed and the "
                    "existing optimizer instance is no longer "
                    "valid. Reinitializing now."
                )
            if gp_priors is not None:
                opt.gp_priors = gp_priors

    if reinitialize and len(X) > 0:
        logger.info(
            f"Importing {len(X)} existing datapoints. " f"This could take a while..."
        )
        opt.tell(
            X,
            y,
            noise_vector=noise,
            gp_burnin=gp_initial_burnin,
            gp_samples=gp_initial_samples,
            n_samples=acq_function_samples,
            progress=True,
        )
        logger.info("Importing finished.")
    return opt


def print_results(
    optimizer: Optimizer,
    result_object: OptimizeResult,
    parameter_names: Sequence[str],
    confidence: float = 0.9,
) -> Tuple[np.ndarray, float, float]:
    """ Log the current results of the optimizer.

    Parameters
    ----------
    optimizer : bask.Optimizer
        Fitted Optimizer object.
    result_object : scipy.optimize.OptimizeResult
        Result object containing the data and the last fitted model.
    parameter_names : Sequence of str
        Names of the parameters to use for printing.
    confidence : float, default=0.9
        Confidence used for the confidence intervals.

    Raises
    ------
    ValueError
        If computation of the optimum was not successful due to numerical reasons.
    """
    logger = logging.getLogger(LOGGER)
    try:
        best_point, best_value = expected_ucb(result_object, alpha=0.0)
        best_point_dict = dict(zip(parameter_names, best_point))
        with optimizer.gp.noise_set_to_zero():
            _, best_std = optimizer.gp.predict(
                optimizer.space.transform([best_point]), return_std=True
            )
        logger.info(f"Current optimum:\n{best_point_dict}")
        estimated_elo = -best_value * 100
        logger.info(
            f"Estimated Elo: {np.around(estimated_elo, 4)} +- "
            f"{np.around(best_std * 100, 4).item()}"
        )
        confidence_mult = confidence_to_mult(confidence)
        lower_bound = np.around(
            -best_value * 100 - confidence_mult * best_std * 100, 4
        ).item()
        upper_bound = np.around(
            -best_value * 100 + confidence_mult * best_std * 100, 4
        ).item()
        logger.info(
            f"{confidence * 100}% confidence interval of the Elo value: "
            f"({lower_bound}, "
            f"{upper_bound})"
        )
        confidence_out = confidence_intervals(
            optimizer=optimizer,
            param_names=parameter_names,
            hdi_prob=confidence,
            opt_samples=1000,
            multimodal=False,
        )
        logger.info(
            f"{confidence * 100}% confidence intervals of the parameters:"
            f"\n{confidence_out}"
        )
        return best_point, estimated_elo, float(best_std * 100)
    except ValueError as e:
        logger.info(
            "Computing current optimum was not successful. "
            "This can happen in rare cases and running the "
            "tuner again usually works."
        )
        raise e


def plot_results(
    optimizer: Optimizer,
    result_object: OptimizeResult,
    iterations: np.ndarray,
    elos: np.ndarray,
    optima: np.ndarray,
    plot_path: str,
    parameter_names: Sequence[str],
    confidence: float = 0.9,
    current_iteration: Optional[int] = None,
) -> None:
    """Plot the current results of the optimizer.

    Parameters
    ----------
    optimizer : bask.Optimizer
        Fitted Optimizer object.
    result_object : scipy.optimize.OptimizeResult
        Result object containing the data and the last fitted model.
    iterations : np.ndarray
        Array containing the iterations at which optima were collected.
    elos : np.ndarray, shape=(n_iterations, 2)
        Array containing the estimated Elo of the optima and the standard error.
    optima : np.ndarray
        Array containing the predicted optimal parameters.
    plot_path : str
        Path to the directory to which the plots should be saved.
    parameter_names : Sequence of str
        Names of the parameters to use for plotting.
    confidence : float
        The confidence level of the normal distribution to plot in the 1d plot.
    current_iteration : int, default=None
        The current iteration of the optimization process.
        If None, the current iteration is assumed to be the amount of points collected.
    """
    logger = logging.getLogger(LOGGER)
    logger.debug("Starting to compute the next plot.")
    timestr = time.strftime("%Y%m%d-%H%M%S")
    dark_gray = "#36393f"

    if current_iteration is None:
        current_iteration = len(optimizer.Xi)

    # First save the landscape:
    save_params = dict()
    if optimizer.space.n_dims == 1:
        fig, ax = plot_objective_1d(
            result=result_object,
            parameter_name=parameter_names[0],
            confidence=confidence,
        )
        save_params["bbox_inches"] = Bbox([[0.5, -0.2], [9.25, 5.5]])
    else:
        plt.style.use("dark_background")
        fig, ax = plt.subplots(
            nrows=optimizer.space.n_dims,
            ncols=optimizer.space.n_dims,
            figsize=(3 * optimizer.space.n_dims, 3 * optimizer.space.n_dims),
        )
        for i in range(optimizer.space.n_dims):
            for j in range(optimizer.space.n_dims):
                ax[i, j].set_facecolor(dark_gray)
        fig.patch.set_facecolor(dark_gray)
        plot_objective(result_object, dimensions=parameter_names, fig=fig, ax=ax)
    plotpath = pathlib.Path(plot_path)
    for subdir in ["landscapes", "elo", "optima"]:
        (plotpath / subdir).mkdir(parents=True, exist_ok=True)
    full_plotpath = plotpath / f"landscapes/landscape-{timestr}-{current_iteration}.png"
    dpi = 150 if optimizer.space.n_dims == 1 else 300
    plt.savefig(full_plotpath, dpi=dpi, facecolor=dark_gray, **save_params)
    logger.info(f"Saving a plot to {full_plotpath}.")
    plt.close(fig)

    # Plot the history of optima:
    fig, ax = plot_optima(
        iterations=iterations,
        optima=optima,
        space=optimizer.space,
        parameter_names=parameter_names,
    )
    full_plotpath = plotpath / f"optima/optima-{timestr}-{current_iteration}.png"
    fig.savefig(full_plotpath, dpi=150, facecolor=dark_gray)
    plt.close(fig)

    # Plot the predicted Elo performance of the optima:
    fig, ax = plot_performance(
        performance=np.hstack([iterations[:, None], elos]), confidence=confidence
    )
    full_plotpath = plotpath / f"elo/elo-{timestr}-{current_iteration}.png"
    fig.savefig(full_plotpath, dpi=150, facecolor=dark_gray)
    plt.close(fig)


def run_match(
    rounds: int = 10,
    engine1_tc: Optional[Union[str, TimeControl]] = None,
    engine2_tc: Optional[Union[str, TimeControl]] = None,
    engine1_st: Optional[Union[str, int]] = None,
    engine2_st: Optional[Union[str, int]] = None,
    engine1_npm: Optional[Union[str, int]] = None,
    engine2_npm: Optional[Union[str, int]] = None,
    engine1_depth: Optional[Union[str, int]] = None,
    engine2_depth: Optional[Union[str, int]] = None,
    engine1_ponder: bool = False,
    engine2_ponder: bool = False,
    engine1_restart: str = "auto",
    engine2_restart: str = "auto",
    timemargin: Optional[Union[str, int]] = None,
    opening_file: Optional[str] = None,
    adjudicate_draws: bool = False,
    draw_movenumber: int = 1,
    draw_movecount: int = 10,
    draw_score: int = 8,
    adjudicate_resign: bool = False,
    resign_movecount: int = 3,
    resign_score: int = 550,
    resign_twosided: bool = False,
    adjudicate_tb: bool = False,
    tb_path: Optional[str] = None,
    concurrency: int = 1,
    debug_mode: bool = False,
    **kwargs: Any,
) -> Iterator[str]:
    """Run a cutechess-cli match of two engines with paired random openings.

    Parameters
    ----------
    rounds : int, default=10
        Number of rounds to play in the match (each round consists of 2 games).
    engine1_tc : str or TimeControl object, default=None
        Time control to use for the first engine. If str, it can be a
        non-increment time control like "10" (10 seconds) or an increment
        time control like "5+1.5" (5 seconds total with 1.5 seconds increment).
        If None, it is assumed that engine1_npm, engine1_st or engine1_depth is
        provided.
    engine2_tc : str or TimeControl object, default=None
        See engine1_tc.
    engine1_st : str or int, default=None
        Time limit in seconds for each move.
        If None, it is assumed that engine1_tc, engine1_npm or engine1_depth is
        provided.
    engine2_st : str or TimeControl object, default=None
        See engine1_tc.
    engine1_npm : str or int, default=None
        Number of nodes per move the engine is allowed to search.
        If None, it is assumed that engine1_tc, engine1_st or engine1_depth is provided.
    engine2_npm : str or int, default=None
        See engine1_npm.
    engine1_depth : str or int, default=None
        Depth the engine is allowed to search.
        If None, it is assumed that engine1_tc, engine1_st or engine1_npm is provided.
    engine2_depth : str or int, default=None
        See engine1_depth.
    engine1_ponder : bool, default=False
        If True, allow engine1 to ponder.
    engine2_ponder : bool, default=False
        See engine1_ponder.
    engine1_restart : str, default="auto"
        Restart mode for engine1. Can be "auto" (default, engine decides), "on" (engine
        is always restarted between games), "off" (engine is never restarted).
    engine2_restart : str, default="auto"
        See engine1_restart.
    timemargin : str or int, default=None
        Allowed number of milliseconds the engines are allowed to go over the time
        limit. If None, the margin is 0.
    opening_file : str, default=None
        Path to the file containing the openings. Can be .epd or .pgn.
        Make sure that the file explicitly has the .epd or .pgn suffix, as it
        is used to detect the format.
    adjudicate_draws : bool, default=False
        Specify, if cutechess-cli is allowed to adjudicate draws, if the
        scores of both engines drop below draw_score for draw_movecount number
        of moves. Only kicks in after draw_movenumber moves have been played.
    draw_movenumber : int, default=1
        Number of moves to play after the opening, before draw adjudication is
        allowed.
    draw_movecount : int, default=10
        Number of moves below the threshold draw_score, without captures and
        pawn moves, before the game is adjudicated as draw.
    draw_score : int, default=8
        Score threshold of the engines in centipawns. If the score of both
        engines drops below this value for draw_movecount consecutive moves,
        and there are no captures and pawn moves, the game is adjudicated as
        draw.
    adjudicate_resign : bool, default=False
        Specify, if cutechess-cli is allowed to adjudicate wins/losses based on
        the engine scores. If one engine’s score drops below -resign_score for
        resign_movecount many moves, the game is considered a loss for this
        engine.
    resign_movecount : int, default=3
        Number of consecutive moves one engine has to output a score below
        the resign_score threshold for the game to be considered a loss for this
        engine.
    resign_score : int, default=550
        Resign score threshold in centipawns. The score of the engine has to
        stay below -resign_score for at least resign_movecount moves for it to
        be adjudicated as a loss.
    resign_twosided : bool, default=False
        If True, the absolute score for both engines has to above resign_score before
        the game is adjudicated.
    adjudicate_tb : bool, default=False
        Allow cutechess-cli to adjudicate games based on Syzygy tablebases.
        If true, tb_path has to be set.
    tb_path : str, default=None
        Path to the folder containing the Syzygy tablebases.
    concurrency : int, default=1
        Number of games to run in parallel. Be careful when running time control
        games, since the engines can negatively impact each other when running
        in parallel.

    Yields
    -------
    out : str
        Results of the cutechess-cli match streamed as str.
    """
    string_array = ["cutechess-cli"]
    string_array.extend(("-concurrency", str(concurrency)))

    if (
        engine1_npm is None
        and engine1_tc is None
        and engine1_st is None
        and engine1_depth is None
    ) or (
        engine2_npm is None
        and engine2_tc is None
        and engine2_st is None
        and engine2_depth is None
    ):
        raise ValueError("A valid time control or nodes configuration is required.")
    string_array.extend(
        _construct_engine_conf(
            id=1,
            engine_npm=engine1_npm,
            engine_tc=engine1_tc,
            engine_st=engine1_st,
            engine_depth=engine1_depth,
            engine_ponder=engine1_ponder,
            engine_restart=engine1_restart,
            timemargin=timemargin,
        )
    )
    string_array.extend(
        _construct_engine_conf(
            id=2,
            engine_npm=engine2_npm,
            engine_tc=engine2_tc,
            engine_st=engine2_st,
            engine_depth=engine2_depth,
            engine_ponder=engine2_ponder,
            engine_restart=engine2_restart,
            timemargin=timemargin,
        )
    )

    if opening_file is None:
        raise ValueError("Providing an opening file is required.")
    opening_path = pathlib.Path(opening_file)
    if not opening_path.exists():
        raise FileNotFoundError(
            f"Opening file the following path was not found: {opening_path}"
        )
    opening_format = opening_path.suffix
    if opening_format not in {".epd", ".pgn"}:
        raise ValueError(
            "Unable to determine opening format. "
            "Make sure to add .epd or .pgn to your filename."
        )
    string_array.extend(
        (
            "-openings",
            f"file={str(opening_path)}",
            f"format={opening_format[1:]}",
            "order=random",
        )
    )

    if adjudicate_draws:
        string_array.extend(
            (
                "-draw",
                f"movenumber={draw_movenumber}",
                f"movecount={draw_movecount}",
                f"score={draw_score}",
            )
        )
    if adjudicate_resign:
        string_array.extend(
            (
                "-resign",
                f"movecount={resign_movecount}",
                f"score={resign_score}",
                f"twosided={str(resign_twosided).lower()}",
            )
        )
    if adjudicate_tb:
        if tb_path is None:
            raise ValueError("No path to tablebases provided.")
        tb_path_object = pathlib.Path(tb_path)
        if not tb_path_object.exists():
            raise FileNotFoundError(
                f"No folder found at the following path: {str(tb_path_object)}"
            )
        string_array.extend(("-tb", str(tb_path_object)))

    string_array.extend(("-rounds", f"{rounds}"))
    string_array.extend(("-games", "2"))
    string_array.append("-repeat")
    string_array.append("-recover")
    string_array.append("-debug")
    string_array.extend(("-pgnout", "out.pgn"))

    with subprocess.Popen(
        string_array, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True
    ) as popen:
        if popen.stdout is not None:
            for line in iter(popen.stdout.readline, ""):
                yield line
        else:
            raise ValueError("No stdout found.")


def is_debug_log(cutechess_line: str,) -> bool:
    """Check if the provided cutechess log line is a debug mode line.

    Parameters
    ----------
    cutechess_line : str
        One line from a cutechess log.

    Returns
    -------
    bool
        True, if the given line is a debug mode line, False otherwise.
    """
    if re.match(r"[0-9]+ [<>]", cutechess_line) is not None:
        return True
    return False


def check_log_for_errors(cutechess_output: List[str],) -> None:
    """Parse the log output produced by cutechess-cli and scan for important errors.

    Parameters
    ----------
    cutechess_output : list of str
        String containing the log output produced by cutechess-cli.
    """
    logger = logging.getLogger(LOGGER)
    for line in cutechess_output:

        # Check for forwarded errors:
        pattern = r"[0-9]+ [<>].+: error (.+)"
        match = re.search(pattern=pattern, string=line)
        if match is not None:
            logger.warning(f"cutechess-cli error: {match.group(1)}")

        # Check for unknown UCI option
        pattern = r"Unknown (?:option|command): (.+)"
        match = re.search(pattern=pattern, string=line)
        if match is not None:
            logger.error(
                f"UCI option {match.group(1)} was unknown to the engine. "
                f"Check if the spelling is correct."
            )
            continue

        # Check for loss on time
        pattern = (
            r"Finished game [0-9]+ \((.+) vs (.+)\): [0-9]-[0-9] {(\S+) "
            r"(?:loses on time)}"
        )
        match = re.search(pattern=pattern, string=line)
        if match is not None:
            engine = match.group(1) if match.group(3) == "White" else match.group(2)
            logger.warning(f"Engine {engine} lost on time as {match.group(3)}.")
            continue

        # Check for connection stall:
        pattern = (
            r"Finished game [0-9]+ \((.+) vs (.+)\): [0-9]-[0-9] {(\S+)'s "
            r"(?:connection stalls)}"
        )
        match = re.search(pattern=pattern, string=line)
        if match is not None:
            engine = match.group(1) if match.group(3) == "White" else match.group(2)
            logger.error(
                f"{engine}'s connection stalled as {match.group(3)}. "
                f"Game result is unreliable."
            )


def parse_experiment_result(
    outstr: str,
    prior_counts: Optional[Sequence[float]] = None,
    n_dirichlet_samples: int = 1000000,
    score_scale: float = 4.0,
    random_state: Union[int, RandomState, None] = None,
    **kwargs: Any,
) -> Tuple[float, float, float]:
    """Parse cutechess-cli result output to extract mean score and error.

    Here we use a simple pentanomial model to exploit paired openings.
    We distinguish the outcomes WW, WD, WL/DD, LD and LL and apply the
    following scoring (note, that the optimizer always minimizes the score):

    +------+------+-------+-----+-----+
    | WW   | WD   | WL/DD | LD  | LL  |
    +======+======+=======+=====+=====+
    | -1.0 | -0.5 | 0.0   | 0.5 | 1.0 |
    +------+------+-------+-----+-----+

    Note: It is important that the match output was produced using
    cutechess-cli using paired openings, otherwise the returned score is
    useless.

    Parameters
    ----------
    output : string (utf-8)
        Match output of cutechess-cli. It assumes the output was coming from
        a head-to-head match with paired openings.
    prior_counts : list-like float or int, default=None
        Pseudo counts to use for WW, WD, WL/DD, LD and LL in the
        pentanomial model.
    n_dirichlet_samples : int, default = 1 000 000
        Number of samples to draw from the Dirichlet distribution in order to
        estimate the standard error of the score.
    score_scale : float, optional (default=4.0)
        Scale of the logistic distribution used to calculate the score. Has to be a
        positive real number
    random_state : int, RandomState instance or None, optional (default: None)
        The generator used to initialize the centers. If int, random_state is
        the seed used by the random number generator; If RandomState instance,
        random_state is the random number generator; If None, the random number
        generator is the RandomState instance used by `np.random`.
    Returns
    -------
    score : float (in [-1, 1])
        Expected (negative) score of the first player (the lower the stronger)
    error : float
        Estimated standard error of the score. Estimated by repeated draws
        from a Dirichlet distribution.
    draw_rate : float
        Estimated draw rate of the match.
    """
    wdl_strings = re.findall(r"Score of.*:\s*([0-9]+\s-\s[0-9]+\s-\s[0-9]+)", outstr)
    array = np.array(
        [np.array([int(y) for y in re.findall(r"[0-9]+", x)]) for x in wdl_strings]
    )
    diffs = np.diff(array, axis=0, prepend=np.array([[0, 0, 0]]))

    # Parse order of finished games to be able to compute the correct pentanomial scores
    finished = np.array(
        [int(x) - 1 for x in re.findall(r"Finished game ([0-9]+)", outstr)]
    )
    diffs = diffs[np.argsort(finished)]

    counts = {"WW": 0, "WD": 0, "WL/DD": 0, "LD": 0, "LL": 0}
    DD = 0  # Track DD separately to compute draw rate
    for i in range(0, len(diffs) - 1, 2):
        match = diffs[i] + diffs[i + 1]
        if match[0] == 2:
            counts["WW"] += 1
        elif match[0] == 1:
            if match[1] == 1:
                counts["WL/DD"] += 1
            else:
                counts["WD"] += 1
        elif match[1] == 1:
            counts["LD"] += 1
        elif match[2] == 2:
            counts["WL/DD"] += 1
            DD += 1
        else:
            counts["LL"] += 1
    counts_array = np.array(list(counts.values()))
    score, error = counts_to_penta(
        counts=counts_array,
        prior_counts=prior_counts,
        n_dirichlet_samples=n_dirichlet_samples,
        score_scale=score_scale,
        random_state=random_state,
        **kwargs,
    )
    draw_rate = (DD + 0.5 * counts["WD"] + 0.5 * counts["LD"] + 1.0) / (
        counts_array.sum() + 3.0
    )
    return score, error, draw_rate


def update_model(
    optimizer: Optimizer,
    point: list,
    score: float,
    variance: float,
    acq_function_samples: int = 1,
    gp_burnin: int = 5,
    gp_samples: int = 300,
    gp_initial_burnin: int = 100,
    gp_initial_samples: int = 300,
) -> None:
    """Update the optimizer model with the newest data.

    Parameters
    ----------
    optimizer : bask.Optimizer
        Optimizer object which is to be updated.
    point : list
        Latest configuration which was tested.
    score : float
        Elo score the configuration achieved.
    variance : float
        Variance of the Elo score of the configuration.
    acq_function_samples : int, default=1
        Number of hyperposterior samples to average the acquisition function over.
    gp_burnin : int, default=5
        Number of burnin iterations to use before keeping samples for the model.
    gp_samples : int, default=300
        Number of samples to collect for the model.
    gp_initial_burnin : int, default=100
        Number of burnin iterations to use for the first initial model fit.
    gp_initial_samples : int, default=300
        Number of samples to collect
    """
    logger = logging.getLogger(LOGGER)
    while True:
        try:
            now = datetime.now()
            # We fetch kwargs manually here to avoid collisions:
            n_samples = acq_function_samples
            gp_burnin = gp_burnin
            gp_samples = gp_samples
            if optimizer.gp.chain_ is None:
                gp_burnin = gp_initial_burnin
                gp_samples = gp_initial_samples
            optimizer.tell(
                x=point,
                y=score,
                noise_vector=variance,
                n_samples=n_samples,
                gp_samples=gp_samples,
                gp_burnin=gp_burnin,
            )
            later = datetime.now()
            difference = (later - now).total_seconds()
            logger.info(f"GP sampling finished ({difference}s)")
            logger.debug(f"GP kernel: {optimizer.gp.kernel_}")
        except ValueError:
            logger.warning(
                "Error encountered during fitting. Trying to sample chain a bit. "
                "If this problem persists, restart the tuner to reinitialize."
            )
            optimizer.gp.sample(n_burnin=11, priors=optimizer.gp_priors)
        else:
            break


def _construct_engine_conf(
    id: int,
    engine_npm: Optional[Union[int, str]] = None,
    engine_tc: Optional[Union[str, TimeControl]] = None,
    engine_st: Optional[Union[int, str]] = None,
    engine_depth: Optional[Union[int, str]] = None,
    engine_ponder: bool = False,
    engine_restart: str = "auto",
    timemargin: Optional[Union[int, str]] = None,
) -> List[str]:
    result = ["-engine", f"conf=engine{id}", f"restart={engine_restart}"]
    if timemargin is not None:
        result.append(f"timemargin={timemargin}")
    if engine_ponder:
        result.append("ponder")
    if engine_npm is not None:
        result.extend(("tc=inf", f"nodes={engine_npm}"))
        return result
    elif engine_st is not None:
        result.append(f"st={str(engine_st)}")
        return result
    elif engine_depth is not None:
        result.extend(("tc=inf", f"depth={str(engine_depth)}"))
        return result
    elif engine_tc is not None:
        if isinstance(engine_tc, str):
            engine_tc = TimeControl.from_string(engine_tc)
        result.append(f"tc={str(engine_tc)}")
        return result
    else:
        raise ValueError(f"No engine time control specified for engine {id}.")
